
from ctypes import *
from numpy.ctypeslib import ndpointer
from numpy import *
import sys
import platform
import time
import random

import numpy as np
import scipy as sp
import scipy.optimize
import cma
from optproblems.base import BoundConstraintsRepair, Individual, Cache, ResourcesExhausted, Problem
from diversipy import maximin_reconstruction, select_greedy_maximin
from dfoalgos.simplex import NelderMeadSimplexSearch, create_regular_simplex


rng = random.Random()
INFINITY = float("inf")


# get library name
dllname = ""
if platform.system() == "Windows":
	dllname = "./bbcomp.dll"
elif platform.system() == "Linux":
	dllname = "./libbbcomp.so"
elif platform.system() == "Darwin":
	dllname = "./libbbcomp.dylib"
else:
	sys.exit("unknown platform")


# initialize dynamic library
bbcomp = CDLL(dllname)
bbcomp.configure.restype = c_int
bbcomp.login.restype = c_int
bbcomp.numberOfTracks.restype = c_int
bbcomp.trackName.restype = c_char_p
bbcomp.setTrack.restype = c_int
bbcomp.numberOfProblems.restype = c_int
bbcomp.setProblem.restype = c_int
bbcomp.dimension.restype = c_int
bbcomp.budget.restype = c_int
bbcomp.evaluations.restype = c_int
bbcomp.evaluate.restype = c_int
bbcomp.evaluate.argtypes = [ndpointer(c_double, flags="C_CONTIGUOUS"), ndpointer(c_double, flags="C_CONTIGUOUS")]
bbcomp.history.restype = c_int
bbcomp.history.argtypes = [c_int, ndpointer(c_double, flags="C_CONTIGUOUS"), ndpointer(c_double, flags="C_CONTIGUOUS")]
bbcomp.errorMessage.restype = c_char_p


# configuration
LOGFILEPATH = "."
USERNAME = "demoaccount"
PASSWORD = "demopassword"
TRACKNAME = "trial"
LOGIN_DELAY_SECONDS = 10
LOCK_DELAY_SECONDS  = 60


# network failure resilient functions
def safeLogin():
	while True:
		result = bbcomp.login(USERNAME, PASSWORD)
		if result != 0:
			return
		msg = bbcomp.errorMessage()
		print("WARNING: login failed: ", msg)
		time.sleep(5 + LOGIN_DELAY_SECONDS * rng.random())
		if msg == "already logged in":
			return



def safeSetTrack():
	while True:
		result = bbcomp.setTrack(TRACKNAME)
		if result != 0:
			return
		print("WARNING: setTrack failed: ", bbcomp.errorMessage())
		safeLogin()



def safeGetNumberOfProblems():
	while True:
		result = bbcomp.numberOfProblems()
		if result != 0:
			return result
		print("WARNING: numberOfProblems failed: ", bbcomp.errorMessage())
		safeSetTrack()



def safeSetProblem(problemID):
	while True:
		result = bbcomp.setProblem(problemID)
		if result != 0:
			return
		msg = bbcomp.errorMessage()
		print("WARNING: setProblem failed: ", msg)
		if len(msg) >= 22 and msg[0:22] == "failed to acquire lock":
			time.sleep(15 + LOCK_DELAY_SECONDS * rng.random())
		else:
			safeSetTrack()



def safeEvaluate(problemID, point):
	while True:
		value = zeros(1)
		result = bbcomp.evaluate(point, value)
		if result != 0:
		    return value[0]
		errorMessage = bbcomp.errorMessage()
		print("WARNING: evaluate failed: ", errorMessage)
		safeSetProblem(problemID)
		usedEvaluations = safeGetEvaluations(problemID)
		budget = safeGetBudget(problemID)
		if usedEvaluations >= 0 and budget != 0 and budget - usedEvaluations <= 0:
			raise ResourcesExhausted(errorMessage)



def safeGetDimension(problemID):
	while True:
		result = bbcomp.dimension()
		if result != 0: return result
		print("WARNING: dimension failed: ", bbcomp.errorMessage())
		safeSetProblem(problemID)



def safeGetBudget(problemID):
	while True:
		result = bbcomp.budget()
		if result != 0: return result
		print("WARNING: budget failed: ", bbcomp.errorMessage())
		safeSetProblem(problemID)



def safeGetEvaluations(problemID):
	while True:
		result = bbcomp.evaluations()
		if result >= 0: return result
		print("WARNING: evaluations failed: " + bbcomp.errorMessage())
		safeSetProblem(problemID)



class DistanceMatrixFunction:
    """General distance function.

    This distance function can handle arbitrary exponents and can optionally
    calculate torus distances. Slightly slower than the specialized
    versions. Special cases ``exponent = 1`` and ``exponent = 2`` correspond
    to Manhattan and Euclidean distance, respectively.

    """
    def __init__(self, exponent=2, max_dists_per_dim=None):
        """Constructor.

        Parameters
        ----------
        exponent : scalar, optional
            The exponent in the distance calculation.
        max_dists_per_dim : array_like, optional
            1-D array of largest possible distance in each dimension.
            Providing these values has the consequence of treating the
            cuboid as a torus. This is useful for eliminating edge effects
            induced by the lack of neighbor points outside the bounds of
            the cuboid.

        """
        self.exponent = exponent
        self.max_dists_per_dim = max_dists_per_dim
        if max_dists_per_dim is not None:
            for max_dist in max_dists_per_dim:
                assert max_dist > 0 or max_dist is None


    def __call__(self, points1, points2):
        """Calculate distance matrix, one column at a time."""
        exponent = self.exponent
        max_dists_per_dim = self.max_dists_per_dim
        num_points1, dim1 = points1.shape
        num_points2, dim2 = points2.shape
        assert dim1 == dim2
        distances = np.zeros((num_points1, num_points2))
        if math.isinf(exponent):
            for i in range(dim1):
                diff = np.abs(points1[:, i, None] - points2[:, i], dtype="d")
                if max_dists_per_dim is not None:
                    np.minimum(max_dists_per_dim[i] - diff, diff, out=diff)
                assert np.all(diff >= 0)
                distances = np.maximum(diff, distances)
        else:
            for i in range(dim1):
                diff = np.abs(points1[:, i, None] - points2[:, i], dtype="d")
                if max_dists_per_dim is not None:
                    np.minimum(max_dists_per_dim[i] - diff, diff, out=diff)
                assert np.all(diff >= 0)
                diff **= exponent
                distances += diff
            # in place root function
            np.power(distances, 1.0 / exponent, out=distances)
        return distances



def sort_key(individual):
    """Sort key for lexicographic sorting with special treatment of None.

    None is replaced with infinity (the worst possible value).

    """
    try:
        iter(individual.objective_values)
        key = []
        for objective in individual.objective_values:
            if objective is None:
                objective = INFINITY
            key.append(objective)
    except TypeError:
        key = individual.objective_values
        if key is None:
            key = INFINITY
    return key



def nbc_srs_model(n, N):
    return 1.144 + 2.288 * (1.0/N) + 0.541 * (1.0/n) * math.log(N)



def nbc_const_two(n, N):
    return 2.0



class NearestBetterClustering:

    def __init__(self, edgeLengthFactor=2.0, usedRules=(1,2), useEdgeLengthsForRule1Threshold=False):
        self.edgeLengthFactor = edgeLengthFactor
        if not 1 in usedRules and not 2 in usedRules:
            raise Exception("you have to choose at least one rule")
        self.useEdgeLengthsForRule1Threshold = useEdgeLengthsForRule1Threshold
        self.usedRules = usedRules
        #self.lexiSorter = LexicographicSorting()
        #self.distanceMatrixFunction = DistanceMatrixFunction(2)


    def isBetterThan(self, ind1, ind2):
        return sort_key(ind1) < sort_key(ind2)


    def distanceTo(self, ind1, ind2):
        """Calculate the euclidean distance between this individual and the other."""
        selfPhenome = ind1.phenome
        otherPhenome = ind2.phenome
        numberVariables = len(selfPhenome)
        assert(numberVariables == len(otherPhenome))
        squaredDist = 0.0
        diff = 0.0
        for i in xrange(numberVariables):
            diff = selfPhenome[i] - otherPhenome[i]
            squaredDist += diff * diff
        return math.sqrt(squaredDist)


    def cluster(self, population):
        # shortcuts
        edges = []
        isMaximizing = False
        INFINITY = float("inf")
        individualsToUpdate = population
        otherIndividuals = population
        numberOthers = len(otherIndividuals)
        # initialization
        minDistances = [INFINITY] * len(individualsToUpdate)
        if len(otherIndividuals) == 0:
            # nothing to do
            return minDistances
        for i, individual in enumerate(individualsToUpdate):
            individual.tempIndex = i
        individualsCopy = list(individualsToUpdate)
        otherCopy = list(otherIndividuals)
        individualsCopy.sort(key=sort_key)
        otherCopy.sort(key=sort_key)
        #self.lexiSorter.sort(individualsCopy, isMaximizing)
        #self.lexiSorter.sort(otherCopy, isMaximizing)
        # calculate distances
        largestIndexBetter = -1
        for individual in individualsCopy:
            index = individual.tempIndex
            # shortcut to save comparisons: the first largestIndexBetter+1
            # solutions are already known to be better than the current
            # individual because both lists are sorted by objective value
            allDistances = [self.distanceTo(individual, otherCopy[j]) for j in xrange(largestIndexBetter + 1)]
            j = largestIndexBetter + 1
            # check remaining part of the others
            while j < numberOthers and self.isBetterThan(otherCopy[j], individual):
                allDistances.append(float(self.distanceTo(individual, otherCopy[j])))
                largestIndexBetter = j
                j += 1
            if len(allDistances) > 0:
                minDistanceIndex = np.argmin(allDistances)
                minDistance = allDistances[minDistanceIndex]
                minDistances[index] = minDistance
                edges.append((individual, minDistance, otherCopy[minDistanceIndex]))
#        for edge in edges:
#            print edge[0], "--", edge[1], "-->", edge[2]
        # determine threshold for rule 1
        if self.useEdgeLengthsForRule1Threshold:
            edgeLengths = [edge[1] for edge in edges]
            avgEdgeLength = np.mean(edgeLengths)
            edgeLengthThreshold = self.edgeLengthFactor * avgEdgeLength
        else:
            distanceMatrixFunction = DistanceMatrixFunction(2)
            points = np.array([individual.phenome for individual in population])
            distanceMatrix = distanceMatrixFunction(points, points)
            distanceMatrix[xrange(len(population)),xrange(len(population))] = np.inf
            nearestNeighborDistances = distanceMatrix.min(axis=1)
            avgNearestNeighborDistance = np.mean(nearestNeighborDistances)
            edgeLengthThreshold = self.edgeLengthFactor * avgNearestNeighborDistance
        #print edgeLengthThreshold
        shortEdges = edges
        if 1 in self.usedRules:
            shortEdges = [edge for edge in edges if edge[1] <= edgeLengthThreshold]
        remainingEdges = set(shortEdges)
        #print(len(remainingEdges))
        if 2 in self.usedRules:
            numberVariables = len(population[0].phenome)
            factorB = (-4.69 * 0.0001 * numberVariables ** 2 + 0.0263 * numberVariables + 3.66 / numberVariables - 0.457) * math.log10(len(population))
            factorB += 7.51 * 0.0001 * numberVariables ** 2 - 0.0421 * numberVariables - 2.26 / numberVariables + 1.83
            rule2CutEdges = []
            for outgoingEdge in edges:
                predecessor = outgoingEdge[0]
                incomingEdges = [edge for edge in edges if edge[2] == predecessor]
                incomingEdgeLengths = [incomingEdge[1] for incomingEdge in incomingEdges]
                if len(incomingEdges) >= 3 and outgoingEdge[1] / np.median(incomingEdgeLengths) > factorB:
                    rule2CutEdges.append(outgoingEdge)
                    #print "cutted an edge"
            remainingEdges.difference_update(rule2CutEdges)
        #print(len(remainingEdges))
        candidates = set(edge[2] for edge in edges)
        #candidates = set(edge[2] for edge in remainingEdges)
        potentialOptima = []
        for candidate in candidates:
            isPotentialOptimum = True
            for remainingEdge in remainingEdges:
                if candidate == remainingEdge[0]:
                    isPotentialOptimum = False
                    break
            if isPotentialOptimum:
                potentialOptima.append(candidate)
        # clean up
        for individual in individualsToUpdate:
            del individual.tempIndex
        return potentialOptima, minDistances
        #return minDistances



def restarted_local_search(problem, local_algo, sampling_algo, started_individuals, local_opt_approximations, global_sample_size, archive_mode, seed, repair_component, clust_param_model=None):
    ########################################
    # forced deactivation of global sampling
    global_sample_size = 1
    ########################################
    num_recycled_points = 2 #problem.num_variables #max(len(problem.archive) * 0.5, 10 * problem.num_variables)
    #print "num recycled points", num_recycled_points
    existing_points = [ind.phenome for ind in started_individuals]
    existing_points += [ind.phenome for ind in local_opt_approximations]
    recycling_candidates = []
    for phenome in problem.archive.keys():
        if all(p <= 1.0 for p in phenome) and all(p >= 0.0 for p in phenome):
            recycling_candidates.append(list(phenome))
    recycled_points = select_greedy_maximin(recycling_candidates, min(len(recycling_candidates), num_recycled_points), existing_points=existing_points)
    unused_global_samples = []
    for point in recycled_points:
        obj_value = problem.archive[tuple(point)]
        ind = Individual(phenome=point, objective_values=obj_value)
        unused_global_samples.append(ind)
    counter = 0
    two_stage_iterations = 0
    try:
        while problem.remaining_evaluations > 0:

            start_candidates = []
            if archive_mode == "all" and budget < 6400:
                existing_individuals = []
                for phenome, obj_values in problem.archive.items():
                    existing_individuals.append(Individual(phenome=repair_component(phenome), objective_values=obj_values))
                #existing_individuals = problem.solutions 
            elif archive_mode == "start_points+local_optima":
                existing_individuals = local_opt_approximations + started_individuals
            elif archive_mode == "global_samples+local_optima" or (archive_mode == "all" and budget >= 6400):
                existing_individuals = local_opt_approximations + started_individuals + unused_global_samples

            existing_points = np.atleast_2d([ind.phenome for ind in existing_individuals])
            start_candidates = sampling_algo(global_sample_size, problem.num_variables, existing_points=existing_points)
            start_candidates = [Individual(phenome=point) for point in start_candidates]
            if len(start_candidates) > 1:
                # take the best solution
                problem.batch_evaluate(start_candidates)
                start_candidates.sort(key=sort_key)
                unused_global_samples.extend(start_candidates[1:])
                start_candidates = start_candidates[0:1]
                
            for start_ind in start_candidates:
                tuple_start = tuple(start_ind.phenome)
                relevant_phenomes = list(key for key in problem.archive.keys() if key != tuple_start)
                relevant_phenomes = np.atleast_2d(relevant_phenomes)
                #print relevant_phenomes
                dist = DistanceMatrixFunction(exponent=2, max_dists_per_dim=None)
                distances = dist(np.atleast_2d(start_ind.phenome), relevant_phenomes)
                #print "min dist: ", distances.min()
                step_size = min(0.15, max(0.005, 0.5 * distances.min()))
                started_individuals.append(start_ind)
                opt = local_algo(problem, start_ind.phenome, step_size, do_stop_early=True, seed=seed+counter)
                opt.phenome = repair_component(opt.phenome)
                local_opt_approximations.append(opt)
                counter += 1
                
            two_stage_iterations += 1
    except StopIteration as instance:
        print(instance)
    return local_opt_approximations, counter, two_stage_iterations



def clustering_method(problem, local_algo, sampling_algo, started_individuals, local_opt_approximations, global_sample_size, archive_mode, seed, repair_component, clust_param_model):
    budget = problem.remaining_evaluations
    best_obj_value = float("inf")
    best_ind = None
    num_recycled_points = 2#problem.num_variables #max(len(problem.archive) * 0.5, 10 * problem.num_variables)
    #print "num recycled points", num_recycled_points
    existing_points = [ind.phenome for ind in started_individuals]
    existing_points += [ind.phenome for ind in local_opt_approximations]
    recycling_candidates = []
    for phenome in problem.archive.keys():
        if all(p <= 1.0 for p in phenome) and all(p >= 0.0 for p in phenome):
            recycling_candidates.append(list(phenome))
    recycled_points = select_greedy_maximin(recycling_candidates, num_recycled_points, existing_points=existing_points)
    unused_global_samples = []
    for point in recycled_points:
        #print 0, "recycled", point
        obj_value = problem.archive[tuple(point)]
        ind = Individual(phenome=point, objective_values=obj_value)
        unused_global_samples.append(ind)
    counter = 0
    two_stage_iterations = 0
    try:
        while problem.remaining_evaluations > 0:

            start_candidates = []
            if archive_mode == "all" and budget < 6400:
                existing_individuals = []
                for phenome, obj_values in problem.archive.items():
                    existing_individuals.append(Individual(phenome=repair_component(phenome), objective_values=obj_values))
            elif archive_mode == "start_points+local_optima":
                existing_individuals = local_opt_approximations + started_individuals
            elif archive_mode == "global_samples+local_optima" or (archive_mode == "all" and budget >= 6400):
                existing_individuals = local_opt_approximations + started_individuals + unused_global_samples

            existing_points = np.atleast_2d([ind.phenome for ind in existing_individuals])
            #for ind in existing_individuals:
            #    print counter, "existing", ind.phenome
            start_candidates = sampling_algo(global_sample_size, problem.num_variables, existing_points=existing_points)
            start_candidates = [Individual(phenome=point) for point in start_candidates]
            problem.batch_evaluate(start_candidates)
            #for ind in start_candidates:
            #    print counter, "start cand", ind.phenome
            start_candidates += unused_global_samples
            #print "existing points", len(existing_points)
            #print "start candidates to cluster", len(start_candidates)
            #print "unused global", len(unused_global_samples)
            #for ind in unused_global_samples:
            #    print counter, "unused global", ind.phenome
            
            # clustering
            edge_length_factor = clust_param_model(problem.num_variables, len(start_candidates))
            nearest_better_clustering = NearestBetterClustering(edge_length_factor, usedRules=(1,2))
            selected, histData = nearest_better_clustering.cluster(start_candidates)
            unused_global_samples = []
            for ind in start_candidates:
                if ind not in selected:
                    unused_global_samples.append(ind)
            start_candidates = selected
            
            for start_ind in start_candidates:
                tuple_start = tuple(start_ind.phenome)
                relevant_phenomes = list(key for key in problem.archive.keys() if key != tuple_start)
                relevant_phenomes = np.atleast_2d(relevant_phenomes)
                #print relevant_phenomes
                dist = DistanceMatrixFunction(exponent=2, max_dists_per_dim=None)
                distances = dist(np.atleast_2d(start_ind.phenome), relevant_phenomes)
                #print "min dist: ", distances.min()
                step_size = min(0.15, max(0.005, 0.5 * distances.min()))
                started_individuals.append(start_ind)
                opt = local_algo(problem, start_ind.phenome, step_size, do_stop_early=True, seed=seed+counter)
                opt.phenome = repair_component(opt.phenome)
                local_opt_approximations.append(opt)
                counter += 1
            
            two_stage_iterations += 1
    except StopIteration:
        pass
    return local_opt_approximations, counter, two_stage_iterations



def lbfgsb(problem, start_point):
    num_variables = len(start_point)
    bounds = [(0.0, 1.0) for _ in xrange(num_variables)]
                                                                        #(func, x0, fprime=None, args=(), approx_grad=0, bounds=None, m=10, factr=10000000.0, pgtol=1.0000000000000001e-05, epsilon=1e-08, iprint=-1, maxfun=15000, disp=None)
    try:
        archive = problem.archive
    except AttributeError:
        archive = problem.objective_function.archive
    archive_copy = dict(archive)
    try:
        local_search_result, obj_value, info = sp.optimize.fmin_l_bfgs_b(problem, start_point, fprime=None, approx_grad=True, bounds=bounds, m=10, factr=10000000.0, pgtol=1.0000000000000001e-05, epsilon=1e-08, iprint=-1, maxfun=400000, disp=None)
        ind = Individual(phenome=list(local_search_result), objective_values=obj_value)
        #print objectiveValue, info
    except StopIteration:
        new_individuals = []
        for phenome in archive:
            if phenome not in archive_copy:
                new_individuals.append(Individual(phenome=list(phenome), 
                                                  objective_values=archive[phenome]))
        ind = min(new_individuals, key=sort_key)
    return ind



def cmaes(problem, start_point, step_size, do_stop_early, seed, max_iterations=float("inf")):
    if problem.remaining_evaluations <= 0:
        raise ResourcesExhausted()
    bounds = [problem.min_bounds, problem.max_bounds]
    if do_stop_early:
        tolfun = 1e-3
        tolfunhist = 1e-5
    else:
        tolfun = 0.0
        tolfunhist = 0.0
    archive_copy = dict(problem.archive)
    optionsDict = {'maxfevals': problem.remaining_evaluations,
                    "maxiter": max_iterations,
                    "bounds": bounds,
                    "tolfun": tolfun,
                    "tolfunhist": tolfunhist,
                    "verb_time": False,
                    "verb_log": 0,
                    "verbose": -1,
                    "verb_disp": 0,
                    "seed": seed}
    try:
        # try-block because CMAES might try to slightly exceed the budget
        results = cma.fmin(problem, start_point, step_size, options=optionsDict)
        ind = Individual(phenome=list(results[0]), objective_values=results[1])
    except (StopIteration, ValueError) as e:
        new_individuals = []
        for phenome in problem.archive:
            if phenome not in archive_copy:
                new_individuals.append(Individual(phenome=list(phenome), 
                                                  objective_values=problem.archive[phenome]))
        if len(new_individuals) > 0:
            ind = min(new_individuals, key=sort_key)
        else:
            ind = Individual(phenome=list(start_point), objective_values=problem(start_point))
    print "cmaes", start_point, step_size, problem.consumed_evaluations, ind.objective_values, ind.phenome
    return ind



def nelder_mead(problem, start_point, step_size, do_stop_early, seed=None, max_iterations=float("inf")):
    if problem.remaining_evaluations <= 0:
        raise ResourcesExhausted()
    if do_stop_early:
        xtol = 0.001
        ftol = 0.0001
    else:
        xtol = None
        ftol = None
    init_simplex = create_regular_simplex(start_point, step_size)
    ss = NelderMeadSimplexSearch(problem, init_simplex, max_iterations=max_iterations, xtol=xtol, ftol=ftol, verbosity=0)
    ind = ss.run()
    print "Nelder-Mead", start_point, step_size, problem.consumed_evaluations, ind.objective_values, ind.phenome
    return ind



def final_local_opt(problem, individuals, local_algo, step_size, seed):
    if len(individuals) == 0:
        return
    local_opt_counter = 0
    iteration = 0
    sorted_inds = list(individuals)
    try:
        while True:
            sorted_inds.sort(key=sort_key)
            for i, ind in enumerate(sorted_inds):
                new_ind = local_algo(problem, ind.phenome, step_size * 10 ** -iteration, do_stop_early=False, seed=seed+local_opt_counter)
                local_opt_counter += 1
                print "final opt", ind.objective_values, new_ind.objective_values, problem.consumed_evaluations
                if sort_key(new_ind) < sort_key(ind):
                    sorted_inds[i] = new_ind
            iteration += 1
    except StopIteration:
        pass
    sorted_inds.sort(key=sort_key)
    return sorted_inds



class BBCompProblem(Problem):

    def __init__(self, problemID, **kwargs):
        Problem.__init__(self, self.objFunction, max_evaluations=safeGetBudget(problemID), **kwargs)
        self.numberPreviousEvaluations = safeGetEvaluations(problemID)
        if self.numberPreviousEvaluations > 0:
            print "note that problem", problemID, "has non-empty history (", self.numberPreviousEvaluations, "evaluations )"
        self.num_variables = safeGetDimension(problemID)
        self.problemID = problemID
        self.min_bounds = [0.0] * self.num_variables
        self.max_bounds = [1.0] * self.num_variables
        
        
    def objFunction(self, phenome):
        phenome = np.array(phenome)
        if self.consumed_evaluations <= self.numberPreviousEvaluations:
            storedPhenome = np.zeros(self.num_variables)
            objectiveValue = np.zeros(1)
            result = bbcomp.history(self.consumed_evaluations - 1, storedPhenome, objectiveValue)
            if result == 0:
                # note: this evaluation is lost
                print "WARNING: history failed."
            # sanity check
            elif np.all(phenome == storedPhenome):
                objectiveValue = objectiveValue[0]
                print "successful restoration of evaluation", self.consumed_evaluations, "(", objectiveValue, ")"
                return objectiveValue
            else:
                print "WARNING: replay out of sync with history."
                print self.consumed_evaluations, self.numberPreviousEvaluations
                print phenome
                print storedPhenome
        # new evaluation
        return safeEvaluate(self.problemID, phenome)



# dimension-dependent two-stage algorithm
def solveProblem(problemID):
    # set the problem
    safeSetProblem(problemID)

    # obtain problem properties
    budget = safeGetBudget(problemID)
    dim = safeGetDimension(problemID)
    evals = safeGetEvaluations(problemID)

    bounds = ([0.0] * dim, [1.0] * dim)
    repair_component = BoundConstraintsRepair(bounds, ["reflect"] * dim)
    orig_problem = BBCompProblem(problemID, phenome_preprocessor=repair_component)

    # output status
    if evals == budget:
        print "problem ", problemID, ": already finished"
        return
    elif evals == 0:
        print "problem ", problemID, ": starting from scratch"
    else:
        print "problem ", problemID, ": starting from evaluation ", evals

    #print budget, dim

    base_seed = 13062016
    seed = base_seed + problemID
    random.seed(seed)
    np.random.seed(seed)

    # config
    if dim <= 5:
        global_sample_size = 1
        two_stage_algo = restarted_local_search
        local_algorithm = nelder_mead
    elif dim <= 10:
        global_sample_size = 25 * dim
        two_stage_algo = clustering_method
        local_algorithm = nelder_mead
    elif dim <= 20:
        global_sample_size = 25 * dim
        two_stage_algo = clustering_method
        local_algorithm = cmaes
    else: #if dim >= 32:
        global_sample_size = 1
        two_stage_algo = restarted_local_search
        local_algorithm = cmaes
    archive_mode = "global_samples+local_optima"
    sequential_sampling = maximin_reconstruction
    edge_length_model = nbc_srs_model
    reserved_budget_fraction = 0.1
    reserved_budget = int(budget * reserved_budget_fraction)

    cached_problem = Cache(orig_problem)
    cached_problem.min_bounds = orig_problem.min_bounds
    cached_problem.max_bounds = orig_problem.max_bounds
    cached_problem.num_variables = orig_problem.num_variables
    cached_problem.consumed_evaluations = 0
    started_individuals = []
    local_opt_approximations = []
    # 1) do one BFGS run
    bfgs_budget = min(10 * cached_problem.num_variables ** 2, 100 * cached_problem.num_variables, budget - reserved_budget)
    restricted_problem = Problem(cached_problem, max_evaluations=bfgs_budget)
    cached_problem.remaining_evaluations = bfgs_budget
    bfgs_start_point = [0.5] * cached_problem.num_variables
    bfgs_local_opt = lbfgsb(restricted_problem, bfgs_start_point)
    local_opt_approximations.append(bfgs_local_opt)
    started_individuals.append(Individual(phenome=list(bfgs_start_point), 
                                          objective_values=cached_problem.archive[tuple(bfgs_start_point)]))
    # 2) two-stage algorithm
    cached_problem.remaining_evaluations = budget - reserved_budget - cached_problem.consumed_evaluations
    local_opt_approximations, counter, two_stage_iterations = two_stage_algo(cached_problem, 
                                                                             local_algorithm, 
                                                                             sequential_sampling, 
                                                                             started_individuals,
                                                                             local_opt_approximations,
                                                                             global_sample_size, 
                                                                             archive_mode, 
                                                                             seed, 
                                                                             repair_component,
                                                                             edge_length_model)
    # 3) final local optimization
    cached_problem.remaining_evaluations = reserved_budget
    for opt in local_opt_approximations:
        print(problemID, opt.objective_values, opt.phenome)
    final_step = 1e-5
    if local_algorithm == nelder_mead:
        final_step = 1e-4
    final_optima = final_local_opt(cached_problem, local_opt_approximations, local_algorithm, final_step, seed+counter)
    best_obj_value = final_optima[0].objective_values
    print("")
    for opt in final_optima:
        print(problemID, opt.objective_values, opt.phenome)

    localOptimaFile = open("approximation_sets/found_optima_" + str(problemID) + ".txt", "w")
    for approxOptimum in final_optima:
        stringParts = []
        stringParts.append(approxOptimum.objective_values)
        stringParts.extend(approxOptimum.phenome)
        optimumString = "\t".join(map(str, stringParts))
        localOptimaFile.write(optimumString)
        localOptimaFile.write("\n")
    localOptimaFile.close()



if __name__ == "__main__":

    # setup
    result = bbcomp.configure(1, LOGFILEPATH)
    if result == 0:
        sys.exit("configure() failed: " + bbcomp.errorMessage())

    problem_id = int(sys.argv[1])

    safeLogin()
    safeSetTrack()
    n = safeGetNumberOfProblems()
    
    solveProblem(problem_id)
    exit()

#    # solve all problems in the track
#    for i in range(n):
#        solveProblem(i)

