
from ctypes import *
from numpy.ctypeslib import ndpointer
from numpy import zeros
import sys
import platform
import time
import random
import copy
import math

import numpy as np
import scipy as sp
import scipy.optimize
import cma
from optproblems.base import BoundConstraintsRepair, Individual, Cache, ResourcesExhausted, Problem
from diversipy import maximin_reconstruction, select_greedy_maximin, psa_partition
from dfoalgos.simplex import NelderMeadSimplexSearch, create_regular_simplex
from evoalgos.algo import SMSEMOA
from evoalgos.individual import ESIndividual
from evoalgos.sorting import NonDominatedSorting


rng = random.Random()
INFINITY = float("inf")


# get library name
dllname = ""
if platform.system() == "Windows":
	dllname = "./bbcomp.dll"
elif platform.system() == "Linux":
	dllname = "./libbbcomp.so"
elif platform.system() == "Darwin":
	dllname = "./libbbcomp.dylib"
else:
	sys.exit("unknown platform")


# initialize dynamic library
bbcomp = CDLL(dllname)
bbcomp.configure.restype = c_int
bbcomp.login.restype = c_int
bbcomp.numberOfTracks.restype = c_int
bbcomp.trackName.restype = c_char_p
bbcomp.setTrack.restype = c_int
bbcomp.numberOfProblems.restype = c_int
bbcomp.setProblem.restype = c_int
bbcomp.dimension.restype = c_int
bbcomp.budget.restype = c_int
bbcomp.evaluations.restype = c_int
bbcomp.evaluate.restype = c_int
bbcomp.evaluate.argtypes = [ndpointer(c_double, flags="C_CONTIGUOUS"), ndpointer(c_double, flags="C_CONTIGUOUS")]
bbcomp.history.restype = c_int
bbcomp.history.argtypes = [c_int, ndpointer(c_double, flags="C_CONTIGUOUS"), ndpointer(c_double, flags="C_CONTIGUOUS")]
bbcomp.errorMessage.restype = c_char_p


# configuration
LOGFILEPATH = "."
USERNAME = "demoaccount"
PASSWORD = "demopassword"
TRACKNAME = "trial"
LOGIN_DELAY_SECONDS = 10
LOCK_DELAY_SECONDS  = 60



# network failure resilient functions
def safeLogin():
	while True:
		result = bbcomp.login(USERNAME, PASSWORD)
		if result != 0:
			return
		msg = bbcomp.errorMessage()
		print("WARNING: login failed: ", msg)
		time.sleep(5 + LOGIN_DELAY_SECONDS * rng.random())
		if msg == "already logged in":
			return



def safeSetTrack():
	while True:
		result = bbcomp.setTrack(TRACKNAME)
		if result != 0:
			return
		print("WARNING: setTrack failed: ", bbcomp.errorMessage())
		safeLogin()



def safeGetNumberOfProblems():
	while True:
		result = bbcomp.numberOfProblems()
		if result != 0:
			return result
		print("WARNING: numberOfProblems failed: ", bbcomp.errorMessage())
		safeSetTrack()



def safeSetProblem(problemID):
	while True:
		result = bbcomp.setProblem(problemID)
		if result != 0:
			return
		msg = bbcomp.errorMessage()
		print("WARNING: setProblem failed: ", msg)
		if len(msg) >= 22 and msg[0:22] == "failed to acquire lock":
			time.sleep(15 + LOCK_DELAY_SECONDS * rng.random())
		else:
			safeSetTrack()



def safeEvaluate(problemID, point):
	while True:
		value = zeros(safeGetNumberOfObjectives(problemID))
		result = bbcomp.evaluate(point, value)
		if result != 0:
		    #print value
		    return value.tolist()
		errorMessage = bbcomp.errorMessage()
		print("WARNING: evaluate failed: ", errorMessage)
		safeSetProblem(problemID)
		usedEvaluations = safeGetEvaluations(problemID)
		budget = safeGetBudget(problemID)
		if usedEvaluations >= 0 and budget != 0 and budget - usedEvaluations <= 0:
			raise ResourcesExhausted(errorMessage)



def safeGetDimension(problemID):
	while True:
		result = bbcomp.dimension()
		if result != 0: return result
		print("WARNING: dimension failed: ", bbcomp.errorMessage())
		safeSetProblem(problemID)



def safeGetNumberOfObjectives(problemID):
	while True:
		result = bbcomp.numberOfObjectives()
		if result != 0: return result
		print("WARNING: numberOfObjectives failed: ", bbcomp.errorMessage())
		safeSetProblem(problemID)



def safeGetBudget(problemID):
	while True:
		result = bbcomp.budget()
		if result != 0: return result
		print("WARNING: budget failed: ", bbcomp.errorMessage())
		safeSetProblem(problemID)



def safeGetEvaluations(problemID):
	while True:
		result = bbcomp.evaluations()
		if result >= 0: return result
		print("WARNING: evaluations failed: " + bbcomp.errorMessage())
		safeSetProblem(problemID)



class DistanceMatrixFunction:
    """General distance function.

    This distance function can handle arbitrary exponents and can optionally
    calculate torus distances. Slightly slower than the specialized
    versions. Special cases ``exponent = 1`` and ``exponent = 2`` correspond
    to Manhattan and Euclidean distance, respectively.

    """
    def __init__(self, exponent=2, max_dists_per_dim=None):
        """Constructor.

        Parameters
        ----------
        exponent : scalar, optional
            The exponent in the distance calculation.
        max_dists_per_dim : array_like, optional
            1-D array of largest possible distance in each dimension.
            Providing these values has the consequence of treating the
            cuboid as a torus. This is useful for eliminating edge effects
            induced by the lack of neighbor points outside the bounds of
            the cuboid.

        """
        self.exponent = exponent
        self.max_dists_per_dim = max_dists_per_dim
        if max_dists_per_dim is not None:
            for max_dist in max_dists_per_dim:
                assert max_dist > 0 or max_dist is None


    def __call__(self, points1, points2):
        """Calculate distance matrix, one column at a time."""
        exponent = self.exponent
        max_dists_per_dim = self.max_dists_per_dim
        num_points1, dim1 = points1.shape
        num_points2, dim2 = points2.shape
        assert dim1 == dim2
        distances = np.zeros((num_points1, num_points2))
        if math.isinf(exponent):
            for i in range(dim1):
                diff = np.abs(points1[:, i, None] - points2[:, i], dtype="d")
                if max_dists_per_dim is not None:
                    np.minimum(max_dists_per_dim[i] - diff, diff, out=diff)
                assert np.all(diff >= 0)
                distances = np.maximum(diff, distances)
        else:
            for i in range(dim1):
                diff = np.abs(points1[:, i, None] - points2[:, i], dtype="d")
                if max_dists_per_dim is not None:
                    np.minimum(max_dists_per_dim[i] - diff, diff, out=diff)
                assert np.all(diff >= 0)
                diff **= exponent
                distances += diff
            # in place root function
            np.power(distances, 1.0 / exponent, out=distances)
        return distances



def sort_key(individual):
    """Sort key for lexicographic sorting with special treatment of None.

    None is replaced with infinity (the worst possible value).

    """
    try:
        iter(individual.objective_values)
        key = []
        for objective in individual.objective_values:
            if objective is None:
                objective = INFINITY
            key.append(objective)
    except TypeError:
        key = individual.objective_values
        if key is None:
            key = INFINITY
    return key



def alternating_one_dim_opt(problem, local_algo, sampling_algo, started_individuals, local_opt_approximations, global_sample_size, archive_mode, seed, repair_component, clust_param_model=None):
    ########################################
    # forced deactivation of global sampling
    global_sample_size = 1
    ########################################
    budget = problem.remaining_evaluations
    counter = 0
    two_stage_iterations = 0
    try:
        while problem.remaining_evaluations > 0:

            start_candidates = []
            if archive_mode == "all" and budget < 6400:
                existing_individuals = []
                for phenome, obj_values in problem.archive.items():
                    existing_individuals.append(Individual(phenome=repair_component(list(phenome)), objective_values=obj_values))
                #existing_individuals = problem.solutions 
            elif archive_mode == "start_points+local_optima":
                existing_individuals = local_opt_approximations + started_individuals
            elif archive_mode == "global_samples+local_optima" or (archive_mode == "all" and budget >= 6400):
                existing_individuals = local_opt_approximations + started_individuals + unused_global_samples

            existing_points = np.atleast_2d([ind.phenome for ind in existing_individuals])
            start_candidates = sampling_algo(global_sample_size, problem.num_variables, existing_points=existing_points)
            start_candidates = [Individual(phenome=point) for point in start_candidates]
            if len(start_candidates) > 1:
                # take the best solution
                problem.batch_evaluate(start_candidates)
                start_candidates.sort(key=sort_key)
                unused_global_samples.extend(start_candidates[1:])
                start_candidates = start_candidates[0:1]
                    
            start_ind = start_candidates[0]
            tuple_start = tuple(start_ind.phenome)
            relevant_phenomes = list(key for key in problem.archive.keys() if key != tuple_start)
            relevant_phenomes = np.atleast_2d(relevant_phenomes)
            #print relevant_phenomes
            if relevant_phenomes.size == 0:
                min_dist = float("inf")
            else:
                dist = DistanceMatrixFunction(exponent=2, max_dists_per_dim=None)
                distances = dist(np.atleast_2d(start_ind.phenome), relevant_phenomes)
                min_dist = distances.min()
            #print "min dist: ", min_dist
            
            obj_order = list(range(problem.real_num_objectives))
            random.shuffle(obj_order)
            for obj_index in obj_order:
                problem.obj_subset = [obj_index]
                
                step_size = min(0.15, max(0.05, 0.5 * min_dist))
                started_individuals.append(start_ind)
                opt = local_algo(problem, start_ind.phenome, step_size, do_stop_early=True, seed=seed+counter)
                opt.phenome = repair_component(opt.phenome)
                local_opt_approximations.append(opt)
                start_ind = opt
                counter += 1
                
            two_stage_iterations += 1
    except StopIteration as instance:
        print(instance)
    return local_opt_approximations, counter, two_stage_iterations



def cmaes(problem, start_point, step_size, do_stop_early, seed, max_iterations=float("inf")):
    if problem.remaining_evaluations <= 0:
        raise ResourcesExhausted()
    bounds = [problem.min_bounds, problem.max_bounds]
    if do_stop_early:
        tolfun = 1e-3
        tolfunhist = 1e-5
    else:
        tolfun = 0.0
        tolfunhist = 0.0
    archive_copy = dict(problem.archive)
    optionsDict = {'maxfevals': problem.remaining_evaluations,
                    "maxiter": max_iterations,
                    "bounds": bounds,
                    "tolfun": tolfun,
                    "tolfunhist": tolfunhist,
                    "verb_time": False,
                    "verb_log": 0,
                    "verbose": -1,
                    "verb_disp": 0,
                    "seed": seed}
    try:
        # try-block because CMAES might try to slightly exceed the budget
        results = cma.fmin(problem, start_point, step_size, options=optionsDict)
        ind = Individual(phenome=list(results[0]), objective_values=results[1])
    except (StopIteration, ValueError) as e:
        new_individuals = []
        for phenome in problem.archive:
            if phenome not in archive_copy:
                new_individuals.append(Individual(phenome=list(phenome), 
                                                  objective_values=problem.archive[phenome]))
        if len(new_individuals) > 0:
            ind = min(new_individuals, key=sort_key)
        else:
            ind = Individual(phenome=list(start_point), objective_values=problem(start_point))
    #print "cmaes", start_point, step_size, problem.consumed_evaluations, ind.objective_values, ind.phenome
    return ind



def nelder_mead(problem, start_point, step_size, do_stop_early, seed=None, max_iterations=float("inf")):
    if problem.remaining_evaluations <= 0:
        raise ResourcesExhausted()
    if do_stop_early:
        xtol = 0.001
        ftol = 0.0001
    else:
        xtol = None
        ftol = None
    init_simplex = create_regular_simplex(start_point, step_size)
    ss = NelderMeadSimplexSearch(problem, init_simplex, max_iterations=max_iterations, xtol=xtol, ftol=ftol, verbosity=0)
    ind = ss.run()
    #print "Nelder-Mead", start_point, step_size, problem.consumed_evaluations, ind.objective_values, ind.phenome
    return ind



class BBCompProblem(Problem):

    def __init__(self, problemID, num_objectives, **kwargs):
        self.real_num_objectives = num_objectives
        Problem.__init__(self, self.objFunction, num_objectives=num_objectives, max_evaluations=safeGetBudget(problemID), **kwargs)
        self.numberPreviousEvaluations = safeGetEvaluations(problemID)
        if self.numberPreviousEvaluations > 0:
            print "note that problem", problemID, "has non-empty history (", self.numberPreviousEvaluations, "evaluations )"
        self.num_variables = safeGetDimension(problemID)
        self.problemID = problemID
        self.min_bounds = [0.0] * self.num_variables
        self.max_bounds = [1.0] * self.num_variables
        self.obj_subset = list(range(num_objectives))
        self.archive = dict()
        self.hash_function = tuple
        
        
    def objFunction(self, phenome):
        phenome = np.array(phenome)
        if self.consumed_evaluations <= self.numberPreviousEvaluations:
            storedPhenome = np.zeros(self.num_variables)
            objective_values = np.zeros(safeGetNumberOfObjectives(self.problemID))
            result = bbcomp.history(self.consumed_evaluations - 1, storedPhenome, objective_values)
            if result == 0:
                # note: this evaluation is lost
                print "WARNING: history failed."
            # sanity check
            elif np.all(phenome == storedPhenome):
                #objectiveValue = objectiveValue[0]
                print "successful restoration of evaluation", self.consumed_evaluations, "(", objective_values, ")"
                key = self.hash_function(phenome)
                with self.lock:
                    self.archive[key] = copy.deepcopy(objective_values)
                objective_values = [objective_values[i] for i in self.obj_subset]
                # disable check
                self.num_objectives = len(objective_values)
                if len(objective_values) == 1:
                    objective_values = objective_values[0]
                return objective_values
            else:
                print "WARNING: replay out of sync with history."
                print self.consumed_evaluations, self.numberPreviousEvaluations
                print phenome
                print storedPhenome
        # new evaluation
        objective_values = safeEvaluate(self.problemID, phenome)
        key = self.hash_function(phenome)
        with self.lock:
            self.archive[key] = copy.deepcopy(objective_values)
        objective_values = [objective_values[i] for i in self.obj_subset]
        # disable check
        self.num_objectives = len(objective_values)
        if len(objective_values) == 1:
            objective_values = objective_values[0]
        return objective_values



# dimension-dependent two-stage algorithm
def solveProblem(problemID):
    # set the problem
    safeSetProblem(problemID)

    # obtain problem properties
    budget = safeGetBudget(problemID)
    dim = safeGetDimension(problemID)
    evals = safeGetEvaluations(problemID)
    num_objectives = safeGetNumberOfObjectives(problemID)

    bounds = ([0.0] * dim, [1.0] * dim)
    repair_component = BoundConstraintsRepair(bounds, ["reflect"] * dim)
    orig_problem = BBCompProblem(problemID, num_objectives, phenome_preprocessor=repair_component)

    orig_problem.obj_subset = [0]

    # output status
    if evals == budget:
        print "problem ", problemID, ": already finished"
        return
    elif evals == 0:
        print "problem ", problemID, ": starting from scratch"
    else:
        print "problem ", problemID, ": starting from evaluation ", evals

    #print budget, dim

    base_seed = 28062016
    seed = base_seed + problemID
    random.seed(seed)
    np.random.seed(seed)

    # config
    if dim <= 10:
        local_algorithm = nelder_mead
    else: #if dim >= 32:
        local_algorithm = cmaes
    global_sample_size = 1
    two_stage_algo = alternating_one_dim_opt
    archive_mode = "start_points+local_optima"
    sequential_sampling = maximin_reconstruction
    edge_length_model = nbc_srs_model
    reserved_budget_fraction = 0.25
    reserved_budget = int(budget * reserved_budget_fraction)

    started_individuals = []
    local_opt_approximations = []
    # 2) two-stage algorithm
    orig_problem.remaining_evaluations = budget - reserved_budget - orig_problem.consumed_evaluations
    local_opt_approximations, counter, two_stage_iterations = two_stage_algo(orig_problem, 
                                                                             local_algorithm, 
                                                                             sequential_sampling, 
                                                                             started_individuals,
                                                                             local_opt_approximations,
                                                                             global_sample_size, 
                                                                             archive_mode, 
                                                                             seed, 
                                                                             repair_component,
                                                                             edge_length_model)
    # 3) final multiobjective optimization
    orig_problem.remaining_evaluations = reserved_budget
    all_solutions = []
    init_sigmas = [0.005] * dim
    learning_param = None
    unit_cube = ([0.0] * dim, [1.0] * dim)
    for key in orig_problem.archive:
        ind = ESIndividual(genome=list(key),
                           objective_values=list(orig_problem.archive[key]),
                           learning_param1=learning_param,
                           learning_param2=learning_param,
                           strategy_params=init_sigmas,
                           recombination_type="arithmetic")
        ind.min_bounds = unit_cube[0]
        ind.max_bounds = unit_cube[1]
        all_solutions.append(ind)
    nds = NonDominatedSorting()
    non_dominated_inds = nds.identify_best_group(all_solutions)
    #print(len(non_dominated_inds))
    #for ind in non_dominated_inds:
    #    print("nds", ind.objective_values, ind.phenome)
    if num_objectives == 3 and len(non_dominated_inds) > 100:
        obj_space_points = [ind.objective_values for ind in non_dominated_inds]
        try:
            clusters = psa_partition(obj_space_points, 100)
            indices = []
            for cluster in clusters:
                indices.append(cluster.obtain_representative_index())
        except AssertionError:
            indices = random.sample(range(len(obj_space_points)), 100)
        non_dominated_inds = [non_dominated_inds[i] for i in indices]
    population_size = min(100, len(non_dominated_inds))
    population = non_dominated_inds
    num_offspring = 5
    orig_problem.obj_subset = list(range(num_objectives))
    ea = SMSEMOA(orig_problem, population, population_size, num_offspring, verbosity=0)
    #print(orig_problem.remaining_evaluations, orig_problem.consumed_evaluations)
    ea.run()
    #print(orig_problem.remaining_evaluations, orig_problem.consumed_evaluations)
    all_solutions = []
    for key in orig_problem.archive:
        ind = ESIndividual(genome=list(key),
                           objective_values=list(orig_problem.archive[key]),
                           learning_param1=learning_param,
                           learning_param2=learning_param,
                           strategy_params=init_sigmas,
                           recombination_type="none")
        ind.min_bounds = unit_cube[0]
        ind.max_bounds = unit_cube[1]
        all_solutions.append(ind)
    non_dominated_inds = nds.identify_best_group(all_solutions)
    #print(len(non_dominated_inds))

    localOptimaFile = open("approximation_sets/found_optima_" + str(problemID) + ".txt", "w")
    for approxOptimum in local_opt_approximations:
        stringParts = []
        stringParts.append(approxOptimum.objective_values)
        stringParts.extend(approxOptimum.phenome)
        optimumString = "\t".join(map(str, stringParts))
        localOptimaFile.write(optimumString)
        localOptimaFile.write("\n")
    localOptimaFile.close()

    localOptimaFile = open("approximation_sets/front_" + str(problemID) + ".txt", "w")
    for approxOptimum in non_dominated_inds:
        stringParts = []
        stringParts.extend(approxOptimum.objective_values)
        stringParts.extend(approxOptimum.phenome)
        optimumString = "\t".join(map(str, stringParts))
        localOptimaFile.write(optimumString)
        localOptimaFile.write("\n")
    localOptimaFile.close()



if __name__ == "__main__":

    # setup
    result = bbcomp.configure(1, LOGFILEPATH)
    if result == 0:
        sys.exit("configure() failed: " + bbcomp.errorMessage())

    problem_id = int(sys.argv[1])

    safeLogin()
    safeSetTrack()
    n = safeGetNumberOfProblems()
    
    solveProblem(problem_id)
    exit()

#    # solve all problems in the track
#    for i in range(n):
#        solveProblem(i)

